import * as mc from '@minecraft/server'
import * as ui from '@minecraft/server-ui'
import * as vanilla from './lib/index.js'
import { Vec3 } from './lib/Vec3'

let cinecam = { enabled: {} }

let manual = new ui.ActionFormData
manual.body('Cinecam adds cinematic cameras from Java Edition. Cinecam uses > as a command prefix. The only command is >cinecam. Arguments: \n launch - starts the cinematic camera \n stop - stops the cinematic camera \n (optional, after launch) Strength - Value from 0.25 - 5. \n (optional, before strength) --nolimit - removes the limit of strength \n (optional, after strength) --walk - makes you able to move while in cinematic mode')
manual.button('Close')

let cineui = new ui.ModalFormData
cineui.title({ translate: "cinecam.ui.title" })
cineui.textField({ translate: 'cinecam.ui.strength' }, '', '1')
cineui.toggle({ translate: 'cinecam.ui.no_limit' }, false)
cineui.toggle({ translate: 'cinecam.ui.walk' }, true)
cineui.toggle({ translate: 'cinecam.ui.walk_with_strength' }, false)
cineui.textField({ translate: 'cinecam.ui.beta_args' }, '', '{}')

function crossProduct(a, b) {
    return {
        x: a.y * b.z - a.z * b.y,
        y: -(a.z * b.x - a.x * b.z),
        z: a.x * b.y - a.y * b.x
    }
}

/**
* @param {mc.Player} player
* @param {mc.Vector3} vec3
*/
function calculateLocalCoordinates(player, vec3) {
    let ploc = player.getHeadLocation()
    let vector3 = Vec3(
        vec3.x,
        vec3.y,
        vec3.z
    )
    let zVec = player.getViewDirection();
    let xVec = Vec3(zVec.z, 0, -zVec.x).normalized;
    let yVec = Vec3.cross(zVec, xVec).normalized;
    let location = Vec3.add(
        Vec3.multiply(xVec, vector3.x),
        Vec3.add(
            Vec3.multiply(yVec, vector3.y),
            Vec3.multiply(zVec, vector3.z)
        )
    );
    //console.warn(typeof location)
    return Vec3.add(Vec3(ploc.x, ploc.y, ploc.z), location);
}

/**
* @param {mc.Player} player
*/
async function CineUI(player) {
    let result = await forceShow(player, cineui);
    if (result.cancelled == true) return;
    let player_name = player.nameTag
    let uiargs = result.formValues;
    let strength = uiargs[0];
    let nolimit = uiargs[1];
    let walk = uiargs[2];
    let walk_with_strength = uiargs[3]
    let beta = uiargs[4];
    let betajson = JSON.parse(beta);
    let betazoom = betajson.zoom;
    let betashift = betajson.shift
    let shift_zoom = "-3.5"
    cinecam.enabled[player_name] = true
    if (strength == '') return
    if ((parseFloat(strength) <= 0.25 || parseFloat(strength) >= 5.00) && !nolimit) {
        player.runCommandAsync('tellraw @s {"rawtext":[{"text":"§cStrength needs to be more or equal to§a 0.25 §cand less or equal to§a 5§r.§§§aBig number = Slow, more cinematic§r. Default value is §a1."}]}')
        return
    }
    if (beta != "{}" && beta != "") {
        if (mc.world.getDynamicProperty("cinecam:beta_args_enabled") == false || mc.world.getDynamicProperty("cinecam:beta_args_enabled") == undefined) {
            player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§c You can't use beta arguments: They are currently disabled in this world."}]}`)
            return;
        }
    }
    if (betashift != undefined && betazoom != undefined) {
        if (mc.world.getDynamicProperty("cinecam:beta_args_enabled") != false) {
            player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§c You can't arguments 'shift' and 'zoom' together"}]}`)
            return;
        }
    }
    player.runCommandAsync(`title @s title §aCineCam started!`)
    let loop = mc.system.runInterval(() => {
        if (betazoom == undefined) {
            if (betashift == undefined) {
                player.camera.setCamera("minecraft:free", {
                    easeOptions: { easeTime: parseInt(strength), easeType: mc.EasingType.Linear }, rotation: player.getRotation(), location: {
                        x: (player.getHeadLocation().x + (!walk_with_strength ? player.getVelocity().x * 16 : 0)),
                        y: (player.getHeadLocation().y + (!walk_with_strength ? player.getVelocity().y * 16 : 0)),
                        z: (player.getHeadLocation().z + (!walk_with_strength ? player.getVelocity().z * 16 : 0))
                    }
                })
                //player.sendMessage(`${player.getVelocity().x} ${player.getVelocity().y} ${player.getVelocity().z} `)

            }
            else {
                //player.runCommandAsync(`camera @s set minecraft:free ease ${strength} linear pos ^${betashift}^1.7^${shift_zoom} rot ~~`)
                player.camera.setCamera(
                    "minecraft:free",
                    {
                        rotation: player.getHeadLocation(),
                        easeOptions: { easeTime: parseInt(strength), easeType: mc.EasingType.Linear },
                        rotation: player.getRotation(),
                        location: {
                            x: (calculateLocalCoordinates(player, { x: betashift.x, y: 0, z: betashift.z }).x + (!walk_with_strength ? player.getVelocity().x * 16 : 0)),
                            y: (calculateLocalCoordinates(player, { x: betashift.x, y: 0, z: betashift.z }).y + (!walk_with_strength ? player.getVelocity().y * 16 : 0)),
                            z: (calculateLocalCoordinates(player, { x: betashift.x, y: 0, z: betashift.z }).z + (!walk_with_strength ? player.getVelocity().z * 16 : 0)),
                        },
                    }
                )
                //console.warn(calculateLocalCoordinates(player, { x: 0, y: 2, z: 1 }))
            }
        }
        else {
            let finalcord = {
                x: (calculateLocalCoordinates(player, { x: 0, y: 0, z: betazoom }).x + player.getVelocity().x * 16),
                y: (calculateLocalCoordinates(player, { x: 0, y: 0, z: betazoom }).y + player.getVelocity().y * 16),
                z: (calculateLocalCoordinates(player, { x: 0, y: 0, z: betazoom }).z + player.getVelocity().z * 16),
            }
            for (let i = 0; i <= betazoom; i++) {
                let locord = {
                    x: (calculateLocalCoordinates(player, { x: 0, y: 0, z: i - 1 }).x + player.getVelocity().x * 16),
                    y: (calculateLocalCoordinates(player, { x: 0, y: 0, z: i - 1 }).y + player.getVelocity().y * 16),
                    z: (calculateLocalCoordinates(player, { x: 0, y: 0, z: i - 1 }).z + player.getVelocity().z * 16),
                }
                if (player.dimension.getBlock(locord).isAir == false) { break }
                finalcord = locord

                player.camera.setCamera(
                    "minecraft:free",
                    {
                        easeOptions: { easeTime: parseInt(strength), easeType: mc.EasingType.Linear },
                        location: finalcord,
                        rotation: player.getRotation()
                    }
                )
            }
        }
        if (betashift == undefined) player.addEffect("invisibility", 30, { showParticles: false })
        if (!walk) player.addEffect(vanilla, 30, { amplifier: 80, showParticles: false })
        if (cinecam.enabled[player_name] == false) {
            mc.system.clearRun(loop)
            player.camera.clear()
        }
    }, 1)
} 6

/**
 * @template {ActionFormData | MessageFormData | ModalFormData} Form
 * @param {Player} player
 * @param {Form} form
 * @param {number} timeout
 * @returns {Promise<Awaited<ReturnType<Form["show"]>>>}
 */
export async function forceShow(player, form, timeout = Infinity) {
    const startTick = mc.system.currentTick;
    while ((mc.system.currentTick - startTick) < timeout) {
        const response = await /** @type {ReturnType<Form["show"]>} */(form.show(player));
        if (response.cancelationReason !== ui.FormCancelationReason.UserBusy) {
            return response;
        }
    };
    throw new Error(`Timed out after ${timeout} ticks`);
};

mc.world.beforeEvents.chatSend.subscribe(eventData => {
    let player = eventData.sender
    let player_name = eventData.sender.nameTag
    let message = eventData.message
    let args = eventData.message.split(' ')
    if (message[0] == '>') {
        if (args[0] == '>cinecam') {
            eventData.cancel = true
            if (args[1] == 'launch') {
                if (args.length > 5) {
                    player.runCommandAsync('tellraw @s {"rawtext":[{"text":"§cToo many arguments!"}]}')
                    return
                }
                let speed = '1'
                let zoom_factor = '0'
                let able_to_walk = false
                if (args[2] == undefined) {
                    speed = '1'
                }
                else {
                    if (args[2] == '--nolimit') {
                        speed = parseFloat(args[3])
                        if (args[4] == '--walk' && args[4] != undefined) {
                            able_to_walk = true
                        }
                    }
                    else {
                        speed = parseFloat(args[2])
                        if (!(speed >= 0.25 && speed <= 5.00)) {
                            player.runCommandAsync('tellraw @s {"rawtext":[{"text":"§cStrength needs to be more or equal to§a 0.25 §cand less or equal to§a 5§r.§§§aBig number = Slow, more cinematic§r. Default value is §a1."}]}')
                            return
                        }
                        if (args[3] == '--walk' && args[3] != undefined) {
                            able_to_walk = true
                        }
                    }
                }
                //                if (args[3] == '--beta' && args[3] != undefined) {
                //                    if (args[4] == undefined) {
                //                        player.runCommandAsync('tellraw @s {"rawtext":[{"text":"§cNot enough arguments!"}]}')
                //                        return
                //                    }
                //                }
                //                else {
                //                    player.runCommandAsync('tellraw @s {"rawtext":[{"text":"§cUnknown argument!"}]}')
                //               }
                cinecam.enabled[player_name] = true
                player.runCommandAsync(`title @s title §aCineCam started!`)
                let loop = mc.system.runInterval(() => {
                    player.runCommandAsync(`camera @s set minecraft:free ease ${speed} linear pos ~~1.7~ rot ~~`)
                    player.runCommandAsync(`effect @s invisibility 1 1 true`)
                    if (!able_to_walk) player.runCommandAsync(`effect @s slowness 1 8 true`)
                    if (cinecam.enabled[player_name] == false) {
                        mc.system.clearRun(loop)

                        player.runCommandAsync(`camera @s clear`)
                    }
                }, 1)
            }
            if (args[1] == undefined) {
                //player.runCommandAsync('tellraw @s {"rawtext":[{"text":"§cNot enough arguments!"}]}')
                mc.system.run(() => {
                    CineUI(player)
                })
                return
            }
            else if (args[1] != 'stop' && args[1] != 'launch' && args[1] != 'help' && args[1] != undefined && args[1] != 'ui') {
                player.runCommandAsync('tellraw @s {"rawtext":[{"text":"§cUnknown argument!"}]}')
                return
            }

        }
    }
})

mc.world.beforeEvents.chatSend.subscribe(eventData => {
    let player = eventData.sender
    let player_name = eventData.sender.nameTag
    let message = eventData.message
    let args = eventData.message.split(' ')
    if (message[0] == '>') {
        if (args[0] == '>cinecam') {
            eventData.cancel = true
            if (args[1] == 'stop') {
                cinecam.enabled[player_name] = false
                player.runCommandAsync(`title @s title §cStopping...`)
                let cd = { time_left: {} }
                cd.time_left[player_name] = Date.now() + 1000
                let cdloop = mc.system.runInterval(() => {
                    if (cd.time_left[player_name] <= Date.now()) {
                        player.runCommandAsync(`camera @s clear`)
                        player.runCommandAsync(`title @s subtitle §aDone!`)
                        mc.system.clearRun(cdloop)
                    }
                }, 1)
            }
        }
    }

})

mc.world.beforeEvents.chatSend.subscribe(eventData => {
    let player = eventData.sender
    let player_name = eventData.sender.nameTag
    let message = eventData.message
    let args = eventData.message.split(' ')
    if (message[0] == '>') {
        if (args[0] == '>cinecam') {
            eventData.cancel = true
            if (args[1] == 'help') {
                mc.system.run(() => {
                    let cd = { time_left: {} }
                    player.runCommandAsync('tellraw @s {"rawtext":[{"text":"Manual will be open in 5 seconds, close the chat!"}]}')
                    cd.time_left[player_name] = Date.now() + 5000
                    let cdloop = mc.system.runInterval(() => {
                        if (cd.time_left[player_name] <= Date.now()) {
                            manual.show(player)
                            mc.system.clearRun(cdloop)
                        }
                    }, 1)

                })
            }
        }
    }

})



mc.world.beforeEvents.chatSend.subscribe((eventData) => {
    let player = eventData.sender
    let player_name = eventData.sender.nameTag
    let message = eventData.message
    let args = eventData.message.split(' ')
    if (message[0] == '>') {
        eventData.cancel = true
        if (args[0] == '>cinecam') {
            if (args[1] == 'ui') {
                mc.system.run(() => {
                    CineUI(player);
                })
            }
        }
    }
})

mc.world.beforeEvents.chatSend.subscribe(eventData => {
    let player = eventData.sender
    let player_name = eventData.sender.nameTag
    let message = eventData.message
    let args = eventData.message.split(' ')
    if (message[0] == '>') {
        if (args[0] == '>cineadmin') {
            eventData.cancel = true
            if (!player.hasTag("cinecam_admin")) {
                player.runCommandAsync('tellraw @s {"rawtext":[{"text":"§cYou do not have required privilegies to use this command!"}]}')
                return;
            }
            if (args[1] == 'enable_beta_args') {
                mc.system.run(() => {
                    if (args[2] == 'true') {
                        mc.world.setDynamicProperty("cinecam:beta_args_enabled", true)
                        player.runCommandAsync('tellraw @s {"rawtext":[{"text":"§aBeta arguments enabled!"}]}')
                    }
                    if (args[2] == 'false') {
                        mc.world.setDynamicProperty("cinecam:beta_args_enabled", false)
                        player.runCommandAsync('tellraw @s {"rawtext":[{"text":"§c Beta arguments disabled!"}]}')
                    }
                    if (args[2] == undefined) {
                        player.runCommandAsync('tellraw @s {"rawtext":[{"text":"§cNot enough arguments!"}]}')
                    }
                })
            }
            if (args[1] == undefined) {
                player.runCommandAsync('tellraw @s {"rawtext":[{"text":"§cNot enough arguments!"}]}')
            }
            else if (args[1] != 'enable_beta_args') {
                player.runCommandAsync('tellraw @s {"rawtext":[{"text":"§cUnknown argument!"}]}')
            }
        }
        if (args[0] != '>cineadmin' && args[0] != '>cinecam') {
            player.runCommandAsync('tellraw @s {"rawtext":[{"text":"§cUnknown CineCam Command!"}]}')
        }
    }
})